#!/usr/bin/perl -w -*- perl -*-
# Program written by: Jens Heunemann
# (c) Jens Heunemann
#
# DATE 21.Mar.1998
# Version 1.03
#
# This program is protected by Copyright-law.
# Unauthorised use prohibited
# If you desire to use it, contact
# Jens Heunemann
# jens@consol.de

package Template;

use 5.004;

$Template::VERSION='1.03'; # ==> ALSO Update POD!!!

use Exporter ();
use FileHandle ();
use strict;
use Carp;
use constant DEBUG => 0;
use File::Copy qw(move);

use vars qw($AUTOLOAD);

@Template::ISA=qw(Exporter);
@Template::EXPORT=qw();
@Template::EXPORT_OK=qw();
@Template::EXPORT_TAGS=();

*code=\&add_code;
*print=\&template_print;
*values=\&add_values;

1;

=head1 NAME

Template - Module to use powerful template, e.g. for html

=head1 SYNOPSIS

use Template;
my $t=Template->new("file.tpl"); #create new Template-object
$t->print;  # and print the contents

<or>

use Template;
my $t=Template->new;
$t->template_name("abc.tpl"); # template is loadad from file abc.tpl
$t->type("simple");
$t->values("name"=>"Jens",
           "alter"=>27,
          );

$t->print;

########### contents of abc.tpl:
<html>
<body>
Name: [name]<br>
Alter: [alter]
###########

<or>

use Template;
my $t=Template->new(
   -type=>"simple",
   -name=>"abc.tpl",
   -values=>{"name"=>"Jens","alter"=>27,}
   );
$t->print;

########### contents of abc.tpl:
<html>
<body>
Name: [name]<br>
Alter: [alter]
###########

<or>

use Template;
my $t=Template->new("file.tpl");
$t->code(
         "^greeting"=>\&greeting,
         "^time"=>sub{return &createdate(time)},
         "^count"=>$count
);   # set code-refs for embedded functions
$t->print;   # print template

<or>

use Template;
my $t=Template->new("file.tpl");
my @array=( {"name"=>"jens","alter"=>27},
            {"name"=>"matthias","alter"=>15},
          );
$t->code("zeile"=>\@array);
$t->print;

######### contents of file.tpl
<!--perlstart: zeile -->name:[name],alter:[alter]
<!--perlend: zeile -->
#########


=head1 NOTICE

part of the documentation is not really up to date :)

=head1 ABSTRACT

With this module you can create one or more (Text-)Templates. You can
embed html-comments with a special syntax into the Template, which are
replaced by values provided by subroutines you give as
code-references.  This way you can separate the text from your
perl-programm, making both easier to create and maintain.

As an alternative, you can also provide references to hashes and arrays of hashes or arrays of arrays instead of subroutine-references

=head1 DESCRIPTION

Methods:

=item new

=item new ($ )

create new Template-object and provide optional Filename

Parameters:
either just provide a filename or use named parameters:

-name : Filename
-template : Template-text
-code : reference to a hash which contains the parameters of a call to
        the code-method
-values : a reference to a hash which contains plain values
          useful for -type=>simple templates
-type : either simple or advanced
        advanced templates use <!--perl: ...--> and <!--perlstart: ...-->
        simple templates just use [abc] and [def]
        see synopsis for an example

=item template_name ($ )

  provide Filename for Template

  if called without parameter, returns the filename

=item name ($ )

  same as template_name

=item type ($ )

  this can be set to "simple" or anything else.

  templates of type simple mean that substitution is done on [xxx] parts instead of <!--perl: xxx--> parts.

=item real_name ($ )

  Name of the File where the resulting text (after template-replaces)
  is saved. if this is not set, the printing is done to STDOUT

=item template ($ )

  Text for the template, if you don\'t load it from a file
  If you call this sub without parameter, you get the current template-text, if you call it with a parameter,the provided parameter gets the new template-text

=item add_code

=item code (@)

  @ is split into RegExp/Code-pairs, which are added to the Template.
  e.g.:
  my $t=new Template;
  $t->code("regexp1"=>\&sub1,
     "regexp2"=>\&sub2);

=item add_error

   Add Error-Template-name

=item error(errorstring)

   Set internal errorstring

=item error

   Print Error-Template according to add_error and errorstring

=item read

   Read Template from file provided by B<template_name>

=item write

   If no Template-text is provided, B<read> the Template.
   Write processed Template to file provided by B<real_name>.

=item print

=item print(Filehandle)

   Print processed Template to STDOUT or Filehandle

=head1 TEMPLATE

Html-comments of one of the following types are replaced by the values provided
by the code you gave with B<code>:

nearly uptodate:

<!--perl:\s*FUNCTION PARAMETER1 PARAMETER2\s*-->

or

<!--perlstart:\s*FUNCTION PARAMETER1 ...\s*-->
LAST PARAMETER
<!--perlend: FUNCTION-->

FUNCTION is matched against the provided RegExps, matching
LastInFirstToMatch, for the first matching RegExp the corresponding
code-ref is called with the Template-object as first parameter (so the
code can provide error-messages), the name of FUNCTION and
PARAMETER1..x as the other Parameters.

The sub should return undef if an error occurs, and non-undef when successful.
If an error occurs, the according error-Template is printed.

In the SECOND form (with perlstart and perlend) LAST PARAMETER is the
whole string between.

=head1 ADVANCED USAGE

One tricky thing about these templates is how to use a template in a template

eg. lets take an array of hashes, which should be displayed in a table:

@array=( { "name" => "tobi" , "tel" => "123" } ,
         { "name" => "katarina" , "tel" => "456" } ,
         { "name" => "cornelia" , "tel" => "789" } ,
       );

Suppose you have the following template:
<table>
<tr><th>NAME</th><th>TELEFONE</th>
</tr>
<!--perlstart: table-->
<tr>
  <td><!--perl: name--></td>
  <td><!--perl: tel--></td>
</tr>
<!--perlend: table-->
</table>


now, how to process this template using closures:

my $t=Template->new("filename_of_template.tpl");
$t->code("table"=>
  sub {

    $par=$_[2]; # get text of inner template
    my $t=Template->new; # inner template, without filename
    $t->template($par); # fill inner Template with text

    my %data; # this hash MUST be defined before the code-section
    my $erg=""; #this is the string which is returned in the end

    $t->code("."=>   # arbitrary data
      sub { # is processed by this subroutine
        my $name=$_[1]; # name of the section, in this case either
                        # "name" or "tel"
        return $hash{$name}; # just return the contents of the hash
      },
      );

    # now we are doing our loop through our @array:
    foreach my $ref (@array) {

      %data=%$ref; # hier KEIN my %data!!!
      $erg.=$t->process; # each subsequent $t->process uses the momentary
             # content of %data, so each gives back the next tablerow

    }

    return $erg;

  },
  );
$t->print;

this code works in theory, but I'm too tired now to test it, so please
give me some feedback if you try it out and it works, thanks :)

Process WITHOUT Closures (NEW in Version 1.03)

@array=( { "name" => "tobi" , "tel" => "123" } ,
         { "name" => "katarina" , "tel" => "456" } ,
         { "name" => "cornelia" , "tel" => "789" } ,
       );

Suppose you have the following template:
<table>
<tr><th>NAME</th><th>TELEFONE</th>
</tr>
<!--perlstart: table-->
<tr>
  <td>[name]</td>
  <td>[tel]</td>
</tr>
<!--perlend: table-->
</table>

my $t=Template->new("filename_of_template.tpl");
$t->code("table"=>\@array);


Also new:

A subroutine can return an array of hashes, the correspoding
template-section will then be processed in the necessary order by
Template.pm

so this works too:

my $t=Template->new("filename_of_template.tpl");
$t->code("table"=>sub {  return \@array });

=head1 AUTHOR INFORMATION

Copyright 1998, Jens Heunemann. All rights reserved.

=head1 BUGS

This is very alpha, the main thing works, but the interface will probably reworked
a number of times.

=cut

# new Method for Template
# Entweder Parameter als
# -name=>wert,-name2=>wert
# bergeben,
# oder in der Reihenfolge
# von @order

# template: String mit dem Template, entweder mit $t->template("...") erzeugen
# oder mit $t->read einlesen
# real_name: Name des Ziel-files
# template_name: Name des Template-files
# code: ref auf eine Unterroutine, die in $_ den Namen der Funktion kriegt,
#       und als Parameter die Parameter an die Funktion

sub new {
  my $class=shift;
  my $type=ref($class)||$class;
  my $self={"code"=>[],
      "error_templates"=>[],
      "type"=>"advanced",
     };
  bless $self,$type;
  return $self unless @_>0;

  #Named Parameter (mit - davor) oder nur ein Parameter?
  if (substr($_[0],0,1) eq "-") {
    my ($key,$wert);
    while (scalar @_>1) {
      $key=shift;
      $key =~ s/^-// or croak "Named Parameter for Template must begin with a minus";
#      croak "Named Parameter for Template occured twice" if exists $ {$self}{$key};
      if ($key eq "code") {
  my $code=shift;
  croak "Parameter -code must be a hashreference"
    unless ref($code)=~/HASH/;
  $self->add_code(%{$code});
      }
      elsif ($key eq "name" or $key eq "file") {
  my $name=shift;
  croak "Template $name does not exist or is not readable"
    unless -e $name && -r $name;
  $self->{name}=$name;
      }
      elsif ($key eq "values") {
  my $values_ref=shift;
  croak "Parameter -values must be a hashreference"
    unless ref($values_ref)=~/HASH/;
  $self->add_values($values_ref);
      }
      else {
  $self->{$key}=shift;
      }
    }
    croak "Odd number of Named Parameters for Template" if scalar @_ == 1;
  } else {
    my $name=shift;
    croak "Template $name does not exist or is not readable"
      unless -e $name && -r $name;
    $self->{"name"}=$name;
#      my @order=qw(name);
#      for (0..$#_) {
#        croak "Too many arguments" if $_>$#order;
#        $self->{$order[$_]}=shift;
#      }
  }
  return $self;
}

sub template_name {
  my $self=shift;
  if (@_==0) {
    return $ {$self}{'name'};
  }
  else {
    my $name=shift;
    croak "Template $name does not exist or is not readable"
      unless -e $name && -r $name;
    return $ {$self}{'name'}=$name;
  }
}

sub real_name {
  my $self=shift;
  if (@_==0) {
    return $ {$self}{'real_name'};
  }
  else {
    my $name=shift;
    croak "Template-target or Backupfile is not writable or cannot be created"
      unless ((-e $name and -w $name or
         open(T,">$name") and close(T)) and
        (-e "$name.bak" and -w "$name.bak"
         or open(T,">$name.bak") and close(T)));
    return $ {$self}{'real_name'}=$name;
  }
}

sub template {
  # template setzen
  my $self=shift;
  if (@_==0) {
    return \ ($self->{'template'});
  }
  else {
    return $ {$self}{'template'}=shift;
  }
}

sub AUTOLOAD {

  my %autoload_values=("template"=>1,
           "type"=>1,
           "name"=>1,
           "real_name"=>1,
          );

  my $self=shift;
  my $value=$AUTOLOAD;
  $value=~s/^.*:://;

  return if $value =~/DESTROY/;

  die "Wrong subroutine called" unless exists $autoload_values{$value};

  if (@_==0) {
    return \ ($self->{$value});
  }
  else {
    return $ {$self}{$value}=shift;
  }
}

sub add_code {
  # code wird intern als liste
  # [[regexp1,coderef1],[regexp2,coderef2],...]
  # gespeichert
  my $self=shift;
  while (@_>1) {
    my $regexp=shift;
    my $code=shift;
    if (ref($code)=~/CODE/) {
      unshift @{$self->{"code"}},[$regexp,$code];
    }
    elsif (ref($code)=~/SCALAR|HASH|ARRAY/) {
      unshift @{$self->{"code"}},[$regexp,sub { $code }];
    }
    elsif (ref($code)) {
      croak "Error in set_code, wrong parameters:".ref($code);
    }
    else {
      unshift @{$self->{"code"}},[$regexp,sub { $code }];
    }
#      elsif (ref($code)=~/HASH/) {
#  #      my %hash=%$code;
#        unshift @{$self->{"code"}},[$regexp,
#           sub {
#             my $par=$_[-1];
#  #            print $par;
#             $par=~s/\[(.*?)\]/$$code{$1}/g;
#             $par;
#           }];
#      }
#      elsif (ref($code) =~ /ARRAY/) {
#        if (ref($code->[0]) =~ /HASH/) {
#   unshift @{$self->{"code"}},
#     [$regexp,
#      sub {
#        my $par=$_[-1];
#        my $erg="";
#        for (@$code) {
#          my %hash=%$_;
#          my $tmp=$par;
#          $tmp=~s/\[([^]]*)\]/$hash{$1}/g;
#          $erg.=$tmp;
#        }
#        $erg;
#      },
#     ];
#        }
#        elsif (ref($code->[0]) =~ /ARRAY/) {
#   unshift @{$self->{"code"}},
#     [$regexp,
#      sub {
#        my $par=$_[-1];
#        my $erg="";
#        for (@$code) {
#          my @array=@$_;
#          my $tmp=$par;
#          $tmp=~s/\[([^]]*)\]/$array[$1]/g;
#          $erg.=$tmp;
#        }
#        $erg;
#      },
#     ];
#        }
#        else {
#   die "Referenz auf ein Array muss ein Array von Arrays oder Array von Hashes sein";
#        }
#      }
  }
  if (@_==1) {
    croak "Wrong number of parameters to set_code";
  }
}

sub add_values {

  my $self=shift;

  if (ref($_[0]) =~ /HASH/) {
    my $values_ref=shift;
    for (keys %$values_ref) {
      $self->add_code("^\Q$_\E\$"=>sub { return $values_ref->{$_[1]} },
         );
    }
  }
  else {
    eval {
      my %hash=@_;
      for (keys %hash) {
  $self->add_code("^\Q$_\E\$"=>$hash{$_},
           );
      }
    };
    if ($@) {
      die "Fehler bei der Parameterbergabe beim Aufruf von add_values";
    }
  }

}

sub add_error {
  # Fehlertemplates setzen
  # zu jedem errorcode wird ein templatename gespeichert.
  my $self=shift;
  while (@_>1) {
    my $regexp=shift;
    my $name=shift;
    croak "Error-Template does not exist or is not readable"
      unless -e $name && -r $name;
    unshift @{$self->{"error_templates"}},[$regexp,$name];
  }
  if (@_==1) {
    croak "Wrong number of parameters to set_code";
  }
}

sub error {
  # Mit parameter: Fehlercode setzen
  # Ohne Parameter: Fehlermeldung (aus Template) zurckgeben
  my $self=shift;
  if (@_==0) {
    my $t=new Template;
    my $ref;
  ERROR:
    for $ref (@{$self->{"error_templates"}}) {
      if ($self->{"error"} =~ /$ref->[0]/) {
  $t->template_name($ref->[1]);
  last ERROR;
      }
    }
    $t->print;
    return 1;
  }
  else {
    return $self->{"error"}=shift;
  }
}

# noch unschn, im Moment wird noch das Template komplett eingelesen,
#
sub read {
  my $self=shift;
  if (open TEMPLATE,"<$ {$self}{'name'}") {
    $ {$self}{'template'}="";
    while (<TEMPLATE>) {
      $ {$self}{'template'}.=$_;
    }
    close TEMPLATE;
  }
  else {
    croak "Could not open TEMPLATE: $!";
  }
  return $self;
}

sub write {
  my $self=shift;
  $self->read() unless defined $ {$self}{'template'};
  my $fh=new FileHandle;
  if ($fh->open(">$ {$self}{'real_name'}.bak")) {
    $self->print($fh) or return;
    $fh->close;
    move("$ {$self}{'real_name'}.bak",$ {$self}{'real_name'}) or croak "Error while moving $ {$self}{'real_name'}.bak to $ {$self}{'real_name'}: $!";
  }
  else {
    croak "Error while opening $ {$self}{'real_name'}.bak for writing: $!";
  }
  return $self;
}

sub template_print {
  my $self=shift;
  my $fh;
  if (@_>0) {
    $fh=shift;
  }
  else {
    $fh=\*STDOUT;
  }
  print $fh $self->process;
  return $self;
}

sub process {
  my $self=shift;
  my $ref;
  my $regexp;
  my $pos=0;
  my $ergebnis="";
  $self->read() unless defined $ {$self}{'template'};

  if ($self->{type} eq "simple") {
    while ($self->{template} =~
	   m/\G(.*?) #Klammer1
       \[([^]]*)\]/igxs
	  ) {
      $pos=pos($self->{"template"});
#    print $fh $1;
      $ergebnis.=$1;
      my $name=$2;
      my $found=0;
    REGEXP:
      for $ref (@{$self->{"code"}}) {
	$name =~ /$ref->[0]/ and do {
	  $found=1;
	  my $value;
	  $value=$ref->[1]->($self,$name);

	  # Keine Rckgabe von der Unterroutine? dann auch nichts ans Ergebnis dranhngen
	  last REGEXP unless defined $value;

	  if (ref($value) =~ /SCALAR/) {
#   print $fh $$value;
	    $ergebnis.=$$value;
	  }
	  else {
#   print $fh $value;
	    $ergebnis.=$value;
	  }
	  last REGEXP
	}
      }
      $ergebnis.="[$name]" unless ($found);
    }
  }
  else {
  MATCH:
    while ($self->{"template"} =~
	   m/\G(.*?) #Klammer1
       <!--perl  #Anfang des bots
       (?:
       :  #Klammer- -  perl: dann bot ohne Inhalt
       [\s\n]*?       # beliebige leerzeichen
       ([a-zA-Z0-9_]+)  #Klammer2 -  Aufzurufende Funktion
       (?:[\s\n]+(?=[^>\s\n])([^>]*?))? #Klammer3 - Evtl noch Parameter dazu
       [\s\n]*?--> # Ende des bots
       |
     start:  #Klammer- - perlstart: dann bot mit Inhalt
       [\s\n]*?
       ([a-zA-Z0-9_]+)  #Klammer4 - Aufzurufende Funktion
       (?:[\s\n]+(?=[^>\s\n])([^>]*?))? #Klammer5 - Evtl noch Parameter dazu
       [\s\n]*? # Hier keine Parameter
       -->
       (.*?)  #Klammer6 - Inhalt
       <!--perlend(?::\s*\4\s*)?-->
      )
       # oder kein bot gefunden, dann bis zum Ende suchen
#   |\Z
       /ixgs) {
      $pos=pos($self->{"template"});
#    print $fh $1;
      $ergebnis.=$1;
      #    my $bot=$1;
      my $name;
      my @par;
      if ($2) {
	$name=$2;
	@par=split " ",($3||"");
	print join "--",@par if DEBUG;
      }
      else {
	$name=$4;
	@par=split " ",($5||"");
	push @par,$6;
#      push @par,$5;
#      print "Parameter:<br>------<br>@par<br>-----<br>" if DEBUG;
      }
#    carp "Calling function: $name with Parameters ".join(" ",@par)."\n" if DEBUG>0;
    REGEXP:
      for $ref (@{$self->{"code"}}) {
	$name =~ /$ref->[0]/ and do {
	  my $value;
	  $value=$ref->[1]->($self,$name,@par);

	  # Keine Rckgabe von der Unterroutine? dann auch nichts ans Ergebnis dranhngen
	  last REGEXP unless defined $value;

	  if (ref($value) =~ /SCALAR/) {
#   print $fh $$value;
	    $ergebnis.=$$value;
	  }
	  elsif (ref($value)=~/HASH/) {
#           my %hash=%$value;
	    my $parameter=$par[-1];
	    $parameter=~s/\[(.*?)\]/defined($value->{$1})?$value->{$1}:""/eg;
	    $ergebnis.=$parameter;
	  }
	  elsif (ref($value) =~ /ARRAY/) {
	    # leeres Array?
	    if (@$value==0) {
	      # do nothing
	    }
	    elsif (ref($value->[0]) =~ /HASH/) {
	      # array von hashes?
	      my $parameter=$par[-1];
	      for (@$value) {
		my $tmp=$parameter;
		$tmp=~s/\[([^]]*)\]/defined($_->{$1})?$_->{$1}:""/ge;
		$ergebnis.=$tmp;
	      }
	    }
	    elsif (ref($value->[0]) =~ /ARRAY/) {
	      #array von arrays?
	      my $parameter=$par[-1];
	      for (@$value) {
		my $tmp=$parameter;
		$tmp=~s/\[([^]]*)\]/$_->[$1]/g;
		$ergebnis.=$tmp;
	      }
	    }
	    else {
	      die "Referenz auf ein Array muss ein Array von Arrays oder Array von Hashes sein";
	    }
	  }
	  else {
#   print $fh $value;
	    $ergebnis.=$value;
	  }
	  last REGEXP
	}
      }
    }
  }

#  print $fh substr($self->{'template'},$pos);
  $ergebnis.=substr($self->{'template'},$pos);
  return $ergebnis;
}
